# -*- coding: utf-8 -*-

from __future__ import unicode_literals

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
import json
import re
import ssl

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.poolmanager import PoolManager

import mygeotab.serializers

try:
    requests.packages.urllib3.disable_warnings()
except:
    pass


class API(object):
    def __init__(self, username, password=None, database=None, session_id=None, server='my.geotab.com'):
        """
        Creates a new instance of this simple Pythonic wrapper for the MyGeotab API.

        :param username: The username used for MyGeotab servers. Usually an email address.
        :param password: The password associated with the username. Optional if `session_id` is provided.
        :param database: The database or company name. Optional as this usually gets resolved upon authentication.
        :param session_id: A session ID, assigned by the server.
        :param server: The server ie. my23.geotab.com. Optional as this usually gets resolved upon authentication.
        :raise Exception: Raises an Exception if a username, or one of the session_id or password is not provided.
        """
        if username is None:
            raise Exception('`username` cannot be None')
        if password is None and session_id is None:
            raise Exception('`password` and `session_id` must not both be None')
        self.credentials = Credentials(username, session_id, database, server, password)
        self.__reauthorize_count = 0

    @staticmethod
    def from_credentials(credentials):
        """
        Returns a new API object from an existing Credentials object

        :param credentials: The existing saved credentials
        :return: A new API object populated with MyGeotab credentials
        """
        return API(username=credentials.username, password=credentials.password,
                   database=credentials.database, session_id=credentials.session_id,
                   server=credentials.server)

    @property
    def _api_url(self):
        """
        Formats the server URL properly in order to query the API.

        :rtype: str
        :return: A valid MyGeotab API request URL
        """
        if not self.credentials.server:
            self.credentials.server = 'my.geotab.com'
        parsed = urlparse(self.credentials.server)
        base_url = parsed.netloc if parsed.netloc else parsed.path
        base_url.replace('/', '')
        return 'https://' + base_url + '/apiv1'

    @property
    def _is_local(self):
        """
        Are calls being made against a local server.

        :rtype: bool
        :return: True if the calls are being made locally
        """
        return any(s in self._api_url for s in ['127.0.0.1', 'localhost'])

    @staticmethod
    def _process(data):
        """
        Processes the returned JSON from the server.

        :param data: The JSON data in dict form
        :return: The result dict
        :raise MyGeotabException: Raises when a server exception was encountered
        """
        if data:
            if 'error' in data:
                raise MyGeotabException(data['error'])
            if 'result' in data:
                return data['result']
            return data
        return None

    def _process_param_names(self, parameters):
        """
        Allows the use of Pythonic-style parameters with underscores instead of camel-case

        :param parameters: The parameters object dict
        :return: The processed parameters
        """
        for param_name in parameters:
            value = parameters[param_name]
            server_param_name = re.sub(r'_(\w)', lambda m: m.group(1).upper(), param_name)
            if isinstance(value, dict):
                value = self._process_param_names(value)
            parameters[server_param_name] = value
            if server_param_name != param_name:
                del parameters[param_name]
        return parameters

    def _query(self, method, parameters):
        """
        Formats and performs the query against the API

        :param method: The method name.
        :param parameters: A dict of parameters to send
        :return: The JSON-decoded result from the server
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        params = dict(id=-1, method=method, params=parameters)
        headers = {'Content-type': 'application/json; charset=UTF-8'}
        with requests.Session() as s:
            s.mount('https://', GeotabHTTPAdapter())
            r = s.post(self._api_url,
                       data=json.dumps(params,
                                       default=mygeotab.serializers.object_serializer),
                       headers=headers, allow_redirects=True, verify=(not self._is_local))
        return self._process(r.json(object_hook=mygeotab.serializers.object_deserializer))

    def call(self, method, **parameters):
        """
        Makes a call to the API.

        :param method: The method name.
        :param parameters: Additional parameters to send (for example, search=dict(id='b123') )
        :return: The JSON result (decoded into a dict) from the server
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        if method is None:
            raise Exception("Must specify a method name")
        if parameters is None:
            parameters = {}
        parameters = self._process_param_names(parameters)
        if self.credentials is None:
            self.authenticate()
        if 'credentials' not in parameters and self.credentials.session_id:
            parameters['credentials'] = self.credentials.get_param()

        try:
            result = self._query(method, parameters)
            if result is not None:
                self.__reauthorize_count = 0
                return result
        except MyGeotabException as exception:
            if exception.name == 'InvalidUserException' and self.__reauthorize_count == 0:
                self.__reauthorize_count += 1
                self.authenticate()
                return self.call(method, **parameters)
            raise
        return None

    def multi_call(self, *calls):
        """
        Performs a multi-call to the API

        :param calls: A list of call 2-tuples with method name and params (for example, ('Get', dict(typeName='Trip')) )
        :return: The JSON result (decoded into a dict) from the server
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        formatted_calls = [dict(method=call[0], params=call[1]) for call in calls]
        return self.call('ExecuteMultiCall', calls=formatted_calls)

    def get(self, type_name, **parameters):
        """
        Gets entities using the API. Shortcut for using call() with the 'Get' method.

        :param type_name: The type of entity
        :param parameters: Additional parameters to send.
        :return: The JSON result (decoded into a dict) from the server
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        return self.call('Get', type_name=type_name, **parameters)

    def search(self, type_name, **parameters):
        """
        Searches for entities using the API. Shortcut for using get() with a search.

        :param type_name: The type of entity
        :param parameters: Additional parameters to send.
        :return: The JSON result (decoded into a dict) from the server
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        if parameters:
            results_limit = parameters.get('resultsLimit', None)
            if results_limit is not None:
                del parameters['resultsLimit']
            parameters = dict(search=parameters)
            return self.call('Get', type_name=type_name, resultsLimit=results_limit, **parameters)
        return self.get(type_name)

    def add(self, type_name, entity):
        """
        Adds an entity using the API. Shortcut for using call() with the 'Add' method.

        :param type_name: The type of entity
        :param entity: The entity to add
        :return: The id of the object added
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        return self.call('Add', type_name=type_name, entity=entity)

    def set(self, type_name, entity):
        """
        Sets an entity using the API. Shortcut for using call() with the 'Set' method.

        :param type_name: The type of entity
        :param entity: The entity to set
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        return self.call('Set', type_name=type_name, entity=entity)

    def remove(self, type_name, entity):
        """
        Removes an entity using the API. Shortcut for using call() with the 'Remove' method.

        :param type_name: The type of entity
        :param entity: The entity to remove
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        return self.call('Remove', type_name=type_name, entity=entity)

    def authenticate(self):
        """
        Authenticates against the API server.

        :return: A Credentials object with a session ID created by the server
        :raise AuthenticationException: Raises if there was an issue with authenticating or logging in
        :raise MyGeotabException: Raises when an exception occurs on the MyGeotab server
        """
        auth_data = dict(database=self.credentials.database, userName=self.credentials.username,
                         password=self.credentials.password)
        auth_data['global'] = True
        try:
            result = self._query('Authenticate', auth_data)
            if result:
                new_server = result['path']
                server = self.credentials.server
                if new_server != 'ThisServer':
                    server = new_server
                c = result['credentials']
                self.credentials = Credentials(c['userName'], c['sessionId'], c['database'],
                                               server)
                return self.credentials
        except MyGeotabException as exception:
            if exception.name == 'InvalidUserException':
                raise AuthenticationException(self.credentials.username,
                                              self.credentials.database,
                                              self.credentials.server)
            raise


class Credentials(object):
    def __init__(self, username, session_id, database, server, password=None):
        """
        Creates a new instance of a MyGeotab credentials object

        :param username: The username used for MyGeotab servers. Usually an email address.
        :param session_id: A session ID, assigned by the server.
        :param database: The database or company name. Optional as this usually gets resolved upon authentication.
        :param server: The server ie. my23.geotab.com. Optional as this usually gets resolved upon authentication.
        :param password: The password associated with the username. Optional if `session_id` is provided.
        """
        self.username = username
        self.session_id = session_id
        self.database = database
        self.server = server
        self.password = password

    def __str__(self):
        return '{0} @ {1}/{2}'.format(self.username, self.server, self.database)

    def get_param(self):
        """
        A simple representation of the credentials object for passing into the API.authenticate() server call

        :return: The simple credentials object for use by API.authenticate()
        """
        return dict(userName=self.username, sessionId=self.session_id, database=self.database)


class MyGeotabException(Exception):
    def __init__(self, full_error):
        """
        Creates a Pythonic exception for server-side exceptions

        :param full_error: The full JSON-decoded error
        """
        self._full_error = full_error
        main_error = full_error['errors'][0]
        self.name = main_error['name']
        self.message = main_error['message']
        self.stack_trace = main_error['stackTrace']

    def __str__(self):
        error_str = '{0}\n{1}'.format(self.name, self.message)
        if self.stack_trace:
            error_str += '\n\nStacktrace:\n{0}'.format(self.stack_trace)
        return error_str


class AuthenticationException(Exception):
    def __init__(self, username, database, server):
        """
        An exception raised on an unsuccessful authentication with the server

        :param username: The username used for MyGeotab servers. Usually an email address.
        :param database: The database or company name.
        :param server: The server ie. my23.geotab.com.
        """
        self.username = username
        self.database = database
        self.server = server

    def __str__(self):
        return 'Cannot authenticate \'{0} @ {1}/{2}\''.format(self.username, self.server,
                                                              self.database)


class GeotabHTTPAdapter(HTTPAdapter):
    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        self.poolmanager = PoolManager(num_pools=connections,
                                       maxsize=maxsize,
                                       block=block,
                                       ssl_version=ssl.PROTOCOL_TLSv1_2,
                                       **pool_kwargs)


__all__ = ['API', 'Credentials', 'MyGeotabException', 'AuthenticationException']



# -*- coding: utf-8 -*-

import abc

from threading import Thread
from time import sleep


class DataFeedListener(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def on_data(self, data):
        """
        Called when rows of data are received

        :param data: A list of data objects
        """
        return

    @abc.abstractmethod
    def on_error(self, error):
        """
        Called when server errors are encountered. Return False to close the stream.

        :rtype: bool
        :param error: The error object
        :return: If True, keep listening. If False, stop the data feed.
        """
        return False


class DataFeed(object):
    def __init__(self, api, listener, type_name, interval, search=None):
        """
        A simple wrapper for the MyGeotab Data Feed. Create a listener that inherits
        from DataFeedListener to pass in.

        :param api: The MyGeotab API object
        :param listener: The custom DataFeedListener object
        :param type_name: The type of entity
        :param interval: The data retrieval interval (in seconds)
        :param search: The search object
        """
        self.api = api
        self.listener = listener
        self.type_name = type_name
        self.interval = interval
        self.search = search
        self.running = False
        self._version = None
        self._thread = None

    def _run(self):
        """
        Runner for the Data Feed

        """
        while self.running:
            try:
                result = self.api.call('GetFeed', type_name=self.type_name, search=self.search,
                                       from_version=self._version)
                self._version = result['toVersion']
                self.listener.on_data(result['data'])
            except Exception as e:
                if self.listener.on_error(e) is False:
                    break
            if not self.running:
                break
            sleep(self.interval)
        self.running = False

    def start(self, async=True):
        """
        Start the Data Feed

        :param async: If True, run in a separate thread
        """
        self.running = True
        if async:
            self._thread = Thread(target=self._run)
            self._thread.start()
        else:
            self._run()



# -*- coding: utf-8 -*-

import os.path

try:
    import ConfigParser as configparser
except ImportError:
    import configparser
import sys

import click

import mygeotab
import mygeotab.api
import mygeotab.dates


class Session(object):
    def __init__(self):
        self.credentials = None

    @staticmethod
    def _get_config_file():
        config_path = click.get_app_dir(mygeotab.__title__)
        if not os.path.exists(config_path):
            os.makedirs(config_path)
        return os.path.join(config_path, 'config.ini')

    @staticmethod
    def _section_name(database):
        return 'session:{0}'.format(database)

    @staticmethod
    def session_names(config):
        if not config:
            return []
        names = []
        section_names = config.sections()
        for name in section_names:
            if ':' in name:
                names.append(name.split(':')[-1])
        return names

    def save(self):
        if not self.credentials:
            return
        config = configparser.ConfigParser()
        config.read(self._get_config_file())
        database = self.credentials.database

        section_name = self._section_name(database)
        if section_name not in config.sections():
            config.add_section(section_name)
        config.set(section_name, 'username', self.credentials.username)
        config.set(section_name, 'session_id', self.credentials.session_id)
        config.set(section_name, 'database', self.credentials.database)
        config.set(section_name, 'server', self.credentials.server)

        with open(self._get_config_file(), 'w') as configfile:
            config.write(configfile)

    def load(self, name=None):
        config = configparser.ConfigParser()
        config.read(self._get_config_file())
        try:
            if name is None:
                sections = config.sections()
                if len(sections) < 1:
                    self.credentials = None
                    return
                section_name = sections[-1]
            else:
                section_name = self._section_name(name)
            username = config.get(section_name, 'username')
            session_id = config.get(section_name, 'session_id')
            database = config.get(section_name, 'database')
            server = config.get(section_name, 'server')
            self.credentials = mygeotab.api.Credentials(username, session_id, database, server)
        except configparser.NoSectionError:
            self.credentials = None
        except configparser.NoOptionError:
            self.credentials = None

    def get_sessions(self):
        config = configparser.ConfigParser()
        config.read(self._get_config_file())
        return self.session_names(config)

    def get_api(self):
        if self.credentials:
            return mygeotab.api.API.from_credentials(self.credentials)
        return None

    def login(self, username, password=None, database=None, server=None):
        if server:
            api = mygeotab.api.API(username=username, password=password, database=database,
                                   server=server)
        else:
            api = mygeotab.api.API(username=username, password=password, database=database)
        self.credentials = api.authenticate()
        self.save()

    def logout(self):
        if self.credentials:
            database = self.credentials.database
            section_name = self._section_name(database)
            config = configparser.ConfigParser()
            config.read(self._get_config_file())
            config.remove_section(section_name)
            with open(self._get_config_file(), 'w') as configfile:
                config.write(configfile)
        self.credentials = None
        sys.exit(0)


def login(session, user, password, database=None, server=None):
    """
    Logs into a MyGeotab server and stores the returned credentials

    :param session: The current Session object
    :param user: The username used for MyGeotab servers. Usually an email address.
    :param password: The password associated with the username. Optional if `session_id` is provided.
    :param database: The database or company name. Optional as this usually gets resolved upon authentication.
    :param server: The server ie. my23.geotab.com. Optional as this usually gets resolved upon authentication.
    """
    if not user:
        user = click.prompt("Username", type=str)
    if not password:
        password = click.prompt("Password", hide_input=True, type=str)
    try:
        with click.progressbar(length=1, label="Logging in...") as bar:
            session.login(user, password, database, server)
            bar.update(1)
        if session.credentials:
            click.echo('Logged in as: %s' % session.credentials)
            session.load(database)
        return session.get_api()
    except mygeotab.api.AuthenticationException:
        click.echo('Incorrect credentials. Please try again.')
        sys.exit(0)


@click.command(help='Lists active sessions')
@click.pass_obj
def sessions(session):
    """
    Shows the current logged in sessions

    :param session: The current Session object
    """
    active_sessions = session.get_sessions()
    if len(active_sessions) == 0:
        click.echo('(No active sessions)')
        return
    for active_session in active_sessions:
        click.echo(active_session)


@click.command(help='Log out from a MyGeotab server')
@click.argument('database', nargs=1, required=False)
@click.pass_obj
def forget(session, database=None):
    """
    Forgets a session from a MyGeotab server by removing the saved credentials

    :param session: The current Session object
    :param database: The database name to log out from
    """
    session.load(database)
    session.logout()


@click.command(help="Launch an interactive MyGeotab console")
@click.argument('database', nargs=1, required=False)
@click.option('--user', '-u')
@click.option('--password', '-p')
@click.option('--server', default=None, help='The server (ie. my4.geotab.com)')
@click.pass_obj
def console(session, database=None, user=None, password=None, server=None):
    """An interactive Python API console for MyGeotab

    If IPython is installed, it will launch an interactive IPython console instead of the built-in Python console. The
    IPython console has numerous advantages over the stock Python console, including: colors, pretty printing,
    command auto-completion, and more.

    By default, all library objects are available as locals in the console, with api

    :param session: The current Session object
    :param database: The database name to open a console to
    :param user: The username used for MyGeotab servers. Usually an email address.
    :param password: The password associated with the username. Optional if `session_id` is provided.
    :param server: The server ie. my23.geotab.com. Optional as this usually gets resolved upon authentication.
    """
    if not session.credentials:
        login(session, user, password, database, server)
    session.load(database)
    api = session.get_api()
    if not api:
        # This DB hasn't been logged into before
        api = login(session, user, password, database, server)
    try:
        api.search('User', name=session.credentials.username)
    except mygeotab.api.AuthenticationException:
        # Credentials expired, try logging in again
        click.echo('Your session has expired. Please login again.')
        api = login(session, user, password, database, server)

    methods = dict(my=api, mygeotab=mygeotab, dates=mygeotab.dates)
    version = 'MyGeotab Console {0} [Python {1}]'.format(mygeotab.__version__,
                                                         sys.version.replace('\n', ''))
    auth_line = ('Logged in as: %s' % session.credentials) if session.credentials else 'Not logged in'
    banner = '\n'.join([version, auth_line])
    try:
        from IPython import embed

        embed(banner1=banner, user_ns=methods)
    except ImportError:
        import code

        code.interact(banner, local=methods)


@click.group()
@click.version_option()
@click.pass_context
def main(ctx):
    """MyGeotab Python SDK command line tools

    You probably want to use the `console` command
    """
    ctx.obj = Session()
    try:
        ctx.obj.load()
    except IOError:
        pass


main.add_command(console)
main.add_command(forget)
main.add_command(sessions)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from datetime import datetime

import pytz
from tzlocal import get_localzone


def format_iso_datetime(dt):
    """
    Formats the given datetime as a UTC-zoned ISO 8601 date string

    :param dt: The datetime object
    :return: The datetime object in 8601 string form
    """
    dt = localize_datetime(dt, pytz.utc)
    return dt.replace(tzinfo=None).isoformat() + 'Z'


def localize_datetime(dt, tz=None):
    """
    Converts a naive or UTC-localized date into the provided timezone or current machine's timezone

    :param dt: The datetime object
    :param tz: The timezone. If blank or None, the user's native timezone is used
    :return: The localized datetime object
    """
    tz = tz or get_localzone()
    if not dt.tzinfo:
        return tz.localize(dt)
    else:
        return dt.astimezone(tz)



# -*- coding: utf-8 -*-

import re

import six
from dateutil import parser

import mygeotab.dates

datetime_regex = re.compile(r'^\d{4}\-\d{2}\-\d{2}')


def object_serializer(obj):
    return mygeotab.dates.format_iso_datetime(obj) if hasattr(obj, 'isoformat') else obj


def object_deserializer(obj):
    for k, v in obj.items():
        if isinstance(v, six.string_types) and datetime_regex.search(v):
            try:
                obj[k] = parser.parse(v)
            except ValueError:
                obj[k] = v
    return obj



# -*- coding: utf-8 -*-
import re
import ast

from setuptools import setup, find_packages


_version_re = re.compile(r'__version__\s+=\s+(.*)')
with open('mygeotab/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

with open('README.rst', 'r') as f:
    readme = f.read()
with open('CHANGELOG.rst', 'r') as f:
    changelog = f.read()

setup(
    name='mygeotab',
    author='Aaron Toth',
    version=version,
    url='https://github.com/geotab/mygeotab-python',
    description='An unofficial Python client for the MyGeotab API',
    long_description=readme + '\n\n' + changelog,
    install_requires=[
        'requests>=2.6',
        'click>=3.3',
        'pytz>=2014.7',
        'python-dateutil>=2.4',
        'six>=1.9',
        'tzlocal>=1.2'
    ],
    extras_require={
        'console': ['ipython']
    },
    test_suite="tests",
    include_package_data=True,
    packages=find_packages(),
    package_data={'': ['LICENSE']},
    package_dir={'mygeotab': 'mygeotab'},
    license='Apache 2.0',
    entry_points='''
        [console_scripts]
        myg=mygeotab.cli:main
    ''',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries'
    ),
)
